

;; This takes care of all of the caching etc
;; and implements:

;; (model-push-hypothesis! h resp) -- pushes the hypothesis h with respones resp onto data
;; (model-set-correct! d) -- a set of data
;; (model-set-data! d)
;; (model-set-which-item! w) -- a list saying which item we are on
;; (model-prediction p) -- takes a parameter value and computes the model predicted proportions (possibly with fancy caching)
;; (get-initial-p) -- what p we can start with?


; http://community.schemewiki.org/?simple-object


; -----------------------------------------------------------------------------------------
;; Define across parameters (outside here, so that they are shared)

(define (set-model)
  (set! across-parameters (vector 1.0 0.0 0.0 0.0 0.0 0.5 0.5 -0.5 -0.5 ))
  (set! across-proposers (vector (lambda (x) (positive-normal-proposal x 0.2)) ;; temperature
                                 (lambda (x) (normal-proposal x 0.1)) ;;intercept
                                 (lambda (x) (normal-proposal x 0.1)) ;; linear
                                 (lambda (x) (normal-proposal x 0.1)) ;; quadratic
                                 (lambda (x) (normal-proposal x 0.1)) ;; third order
                                 (lambda (x) (beta-proposal x 200)) ;; alpha
                                 (lambda (x) (negative-normal-proposal x 0.025))
                                 (lambda (x) (negative-normal-proposal x 0.025)))) ;; baserate
  (set! make-model make-model:ptrue))

; -----------------------------------------------------------------------------------------
(define (make-model:ptrue)
  
  (define META-description "Assume the prior is e to the minus polynomial function of probability of being true." )
  (define META-notes       "(Uses C external library for running-ll)." )
  
  (define hypotheses null) ;; a vector of hypotheses
  (define responses null) ;; a vector of hypotheses
  (define responses_cb null)
  (define true-counts null) ;; for each hypothesis, how many things does it map to true?
  (define correct null) ;; stored in collapsed format
  (define correct_cb null)
  (define which-item null)
  (define which-item_cb null)
  (define len 0)
  
  (define (initialize-model) 0)  
  (define (model-set-data! d) null)
  (define (get-initial-p) (vector))
  (define (get-proposers) (vector))
 
                           
 (define (model-set-correct! c) 
    (set! correct c)
    (set! len (length correct))
    (set! correct_cb (list->cblock (map to01 correct) 'int))
    )
  
  (define (model-set-which-item! w) 
    (set! which-item w)
    (set! which-item_cb (list->cblock which-item 'int))
    )
  
  ;; adds h to the model, with c and l
  (define (model-push-hypothesis! h resp) 
    (set! hypotheses     (cons h hypotheses ))
    (set! responses      (cons resp responses ))
    (set! responses_cb (cons (list->cblock (map to01 resp) 'int) responses_cb))
    (set! true-counts    (cons (length (filter (lambda (x) (eqv? x #t)) resp)) 
                               true-counts))
    )

  (define (clear-hypotheses!) null)
  
  ;; --------------------------
  ;; Local helper functions
  
  
  ;; --------------------------
  (define (model-prediction p gp ap)
    
    (let* ((PARAM-prior-temperature (vector-ref ap 0))
           (PARAM-prior-intercept (vector-ref ap 1))
           (PARAM-prior-linear    (vector-ref ap 2))
           (PARAM-prior-quadratic (vector-ref ap 3))
           (PARAM-prior-thirdorder (vector-ref ap 4))
           (PARAM-alpha             (vector-ref ap 5))
           (PARAM-baserate          (vector-ref ap 6))
           
           (PARAM-llBeta-positive            (vector-ref ap 7))
           (PARAM-llBeta-negative            (vector-ref ap 8))
 
           
           ;; The running LL with the decay
;           (running (map (lambda (rs) (running-ll-SCHEME PARAM-alpha 
;                                                         PARAM-baserate
;                                                         PARAM-llBeta
;                                                         rs
           ;                                                         correct
           ;                                                         which-item))
           ;                         responses))
           ( running (map (lambda (rscb) (running-ll-C PARAM-alpha  ; (a b llb len resp_cb cor_cb which_cb)
                                                    PARAM-baserate
                                                    PARAM-llBeta-positive
                                                    PARAM-llBeta-negative
                                                    len
                                                    rscb
                                                    correct_cb
                                                    which-item_cb))
                          responses_cb))
           
           (prior-score (map (lambda (tc) (let ((tcl (/ tc len)))
                                            (+ PARAM-prior-intercept 
                                               (* tcl PARAM-prior-linear) 
                                               (* tcl tcl PARAM-prior-quadratic)
                                               (* tcl tcl tcl PARAM-prior-thirdorder)
                                               )))
                             true-counts))
           (priors (let ((prior-normalizer (logsumexp-lazy prior-score)))
                     (map (lambda (x) (- x prior-normalizer)) prior-score))))
      
      (map (lambda (response-i running-i) ;; the responses for the i'th data point and the running-ll for the i'th data point
             (let* ((posterior (map + running-i priors ))
                    (normalizer (logsumexp-lazy posterior)))
               (sum (map (lambda (rij postj) 
                           (* (if (eqv? rij #t) 1.0 0.0) 
                              (exp (- postj normalizer))))
                         response-i
                         posterior))))
           (zip-list responses) ;; these in the right format--first list is all hyps on the first data point
           (zip-list running)
           )))
  
    ;; ------------------------------------------------
  ;;; And done, so make this thing a class
  (lambda (message . args)
    (apply (case message 
             ((initialize-model) initialize-model) 
             ((get-initial-p) get-initial-p)
             ((get-proposers) get-proposers) 
             ((model-push-hypothesis!) model-push-hypothesis!) 
             ((model-set-correct!) model-set-correct!) 
             ((model-set-which-item!) model-set-which-item!) 
             ((model-set-data!) model-set-data!)
             ((model-prediction) model-prediction) 
	     ((clear-hypotheses!) clear-hypotheses!)
             (else (die (string-append "Unknown Method! " (symbol->string message))))) 
           args))) 

                       