

;; This takes care of all of the caching etc
;; and implements:

;; (model-push-hypothesis! h resp) -- pushes the hypothesis h with respones resp onto data
;; (model-set-correct! d) -- a set of data
;; (model-set-data! d)
;; (model-set-which-item! w) -- a list saying which item we are on
;; (model-prediction p) -- takes a parameter value and computes the model predicted proportions (possibly with fancy caching)
;; (get-lower-bound) -- retrieve bounds on p
;; (get-upper-bound) 
;; (get-initial-p) -- what p we can start with?

; http://community.schemewiki.org/?simple-object
; -----------------------------------------------------------------------------------------
;; Define across parameters (outside here, so that they are shared)

(define (set-model)
  (set! across-parameters (vector 0.5 0.5 -0.5 -0.5))
  (set! across-proposers (vector (lambda (x) (beta-proposal x 200))
                                 (lambda (x) (beta-proposal x 200))
                                 (lambda (x) (negative-normal-proposal x 0.025))
                                 (lambda (x) (negative-normal-proposal x 0.025))))
  (set! make-model make-model:uniform))

; -----------------------------------------------------------------------------------------
(define (make-model:uniform)
  
  (define META-description "Uniform over hypotheses." )
  (define META-notes       "(Uses C external library for running-ll)." )
  
  (define hypotheses null) ;; a vector of hypotheses
  (define responses null) ;; a vector of hypotheses
  (define responses_cb null)
  (define correct null) ;; stored in collapsed format
  (define correct_cb null) ;; stored in collapsed format
  (define which-item null)
  (define which-item_cb null)
  (define len 0)
  
  (define (initialize-model) 0)  
  (define (model-set-data! d) null)
  (define (get-initial-p) (vector))
  (define (get-proposers) (vector))
  
  (define (model-set-correct! c) 
    (set! correct c)
    (set! correct_cb (list->cblock (map to01 correct) 'int))
    (set! len (length correct))
    )
  
  (define (model-set-which-item! w) 
    (set! which-item w)
    (set! which-item_cb (list->cblock which-item 'int)))
  
  
  ;; adds h to the model, with c and l
  (define (model-push-hypothesis! h resp) 
    (set! hypotheses (cons h hypotheses ))
    (set! responses  (cons resp responses ))
    (set! responses_cb (cons (list->cblock (map to01 resp) 'int) responses_cb))
       )

  (define (clear-hypotheses!) null)

  (define (model-prediction p gp ap)
    
    (let* ((PARAM-alpha             (vector-ref ap 0))
           (PARAM-baserate          (vector-ref ap 1))
           (PARAM-llBeta-positive            (vector-ref ap 2))
           (PARAM-llBeta-negative            (vector-ref ap 3))
           ;; The running LL with the decay
;           (running (map (lambda (rs) (running-ll-SCHEME PARAM-alpha 
;                                                         PARAM-baserate
;                                                         PARAM-llBeta
;                                                         rs
;                                                         correct
;                                                         which-item))
           ;                         responses))
           ( running (map (lambda (rscb) (running-ll-C PARAM-alpha  ; (a b llb len resp_cb cor_cb which_cb)
                                                       PARAM-baserate
                                                       PARAM-llBeta-positive
                                                       PARAM-llBeta-negative
                                                       len
                                                       rscb
                                                       correct_cb
                                                       which-item_cb))
                          responses_cb))
           )
    
    (map (lambda (response-i running-i) ;; the responses for the i'th data point and the running-ll for the i'th data point
           (let* ((posterior running-i)
                  (normalizer (logsumexp-lazy posterior)))
             (sum (map (lambda (rij postj) 
                             (* (if (eqv? rij #t) 1.0 0.0) 
                                (exp (- postj normalizer))))
                           response-i
                           posterior))))
         (zip-list responses) ;; these in the right format--first list is all hyps on the first data point
         (zip-list running)
         )))
  
  (define model-entropy (lambda args 0))
  
  ;; ------------------------------------------------
  ;;; And done, so make this thing a class
  (lambda (message . args)
    (apply (case message 
             ((initialize-model) initialize-model) 
             ((get-initial-p) get-initial-p)
             ((get-lower-bound) get-lower-bound) 
             ((get-upper-bound) get-upper-bound) 
             ((get-proposers) get-proposers) 
             ((model-push-hypothesis!) model-push-hypothesis!) 
             ((model-set-correct!) model-set-correct!) 
             ((model-set-which-item!) model-set-which-item!) 
             ((model-set-data!) model-set-data!)
             ((model-prediction) model-prediction) 
	     ((clear-hypotheses!) clear-hypotheses!)
             ((model-entropy) model-entropy)
             (else (die (string-append "Unknown Method! " (symbol->string message))))) 
           args))) 

                       


